#include <Wire.h>
#include <math.h>
#include <LiquidCrystal.h>

// defines
#define TEMPADDRESS 0x48    // Baseadress first LM75
#define TRAFFICLIGHTS 0x3F  // Expander Traffic lights
#define TEMP 0              // Temp register
#define UNTEN 2             // Lower threshold
#define OBEN 3              // Upper threshold
#define LCD_OFFSET 0        // 4 for 20 x 4 Character

// define thresholds
#define uTEMP 28
#define oTEMP 32
#define uGAS 15
#define mGAS 25
#define oGAS 65

// define INPUTs & OUTPUTs
#define LCD_D7 2
#define LCD_D6 3
#define LCD_D5 4
#define LCD_D4 5
#define SWITCH_1 6
#define SWITCH_2 7
#define F_BUZZER_1 9  // BUZZER Temp
#define F_BUZZER_2 8  // BUZZER Gas
#define L_GAS 10      // FAN Gas
#define L_TEMP 13     // FAN Temp

#define LCD_E 11
#define LCD_RS 12

#define H_TEMP1 A1  // HEATER Temp 1
#define H_TEMP2 A2  // HEATER Temp 2
#define H_TEMP3 A3  // HEATER Temp 3

LiquidCrystal lcd(LCD_RS, LCD_E, LCD_D4, LCD_D5, LCD_D6, LCD_D7);

// ==============================================================
// Functions
int checkStatus(int opt);
void signalGoGreen(int opt);
void signalGoRed(int opt);
int receive_sensor(int opt);
void output(void);
void traffic_signal(int opt);
void buzzerout(void);
int pushS1(int heating);
int pushS2(int heating);

// LM75
double get_LM75_temperature(int device, int regx);
void set_LM75_config(int device, byte value);
byte get_LM75_config(int device);
void set_LM75_threshold(int device, byte regx, double grad);

// Output expander (Input of D/A Converter)
void sendDigitalValue(byte ADDR, byte value);

// global variables
char dataString[7];  // read temperature processed as a string: (-xx)x.x
int status = 0, toChange = true, memStatus = 0, counter_switch = 0, toggle = 0;
unsigned long timestamp = 0, timestampgas = 0, changetones = 0;

void setup() {
  // put your setup code here, to run once:
  Serial.begin(9600);
  while (!Serial)
    ;
  Serial.println("\nProgram transfer complete!");
  Serial.println("Pruefung Sommer 25");
  Serial.println("You can unplug the uno!");
  Serial.end();
  Wire.begin();

  // Set IO-Pins
  for (int i = 2; i < 13; i++) {
    if (i == SWITCH_1 || i == SWITCH_2) {
      pinMode(i, INPUT);
    } else {
      pinMode(i, OUTPUT);
      digitalWrite(i, LOW);
    }
  }
  pinMode(H_TEMP1, OUTPUT);
  pinMode(H_TEMP2, OUTPUT);
  pinMode(H_TEMP3, OUTPUT);
  for (int i = 0; i < 3; i++) {
    set_LM75_threshold(i, OBEN, oTEMP);
    set_LM75_threshold(i, UNTEN, uTEMP);
  }

  lcd.begin(16, 4);
  lcd.clear();
  lcd.home();
  lcd.print("EGS Sommer 25");
  delay(1000);
  lcd.clear();
  signalGoGreen(0);
  output();
}

void loop() {
  int traffic = 0x00;
  while (1) {
    counter_switch = pushS1(pushS2(counter_switch));
    for (int i = 0; i < 4; i++, buzzerout())
      receive_sensor(i);
    traffic = checkStatus(traffic);
  }
}


//Functions
void buzzerout(void) {
  if (!memStatus) {
    delay(20);
    digitalWrite(F_BUZZER_1, LOW);
    digitalWrite(F_BUZZER_2, LOW);
    return;
  }
 
  changetones = (!changetones ? millis() : changetones);
  if (changetones < millis() - 500) {
    changetones = 0;
    toggle = !toggle;
  }
  if ((status & 0b00000111) || bitRead(status, 4))
    digitalWrite(F_BUZZER_1, ((bitRead(status, 4) || bitRead(status, 3)) ? toggle : HIGH));
  else
    digitalWrite(F_BUZZER_1, LOW);
  if (bitRead(status, 4) || bitRead(status, 3))
    digitalWrite(F_BUZZER_2, ((status & 0b00000111) || bitRead(status, 4) ? !toggle : HIGH));
  else
    digitalWrite(F_BUZZER_2, LOW);
  return;
}

int checkStatus(int opt) {
  if (toChange) {
    toChange = false;
    if (bitRead(status, 4)) {
      if (!opt) signalGoRed(0);
      else if (opt == 0x0F) signalGoRed(4);
      else if (opt == 0xF0) signalGoRed(2);
      return 0xFF;
    } else if (bitRead(status, 5)) {
      if (opt == 0x00) signalGoRed(1);
      else if (opt == 0xFF) signalGoGreen(1);
      return 0x0F;
    } else if (!status) {
      if (opt == 0xFF) signalGoGreen(0);
      else if (opt == 0x0F) signalGoGreen(1);
      else if (opt == 0xF0) signalGoGreen(3);
      return 0;
    }
  } else if (opt == 0x0F && bitRead(status, 5)) {
    signalGoRed(4);
    signalGoGreen(2);
    return 0xF0;
  } else if (opt == 0xF0 && bitRead(status, 5)) {
    signalGoRed(2);
    signalGoGreen(4);
    return 0x0F;
  }
  return opt;
}

void output(void) {
  lcd.home();
  lcd.print("Temp1 00\337C    OK");
  lcd.setCursor(0, 1);
  lcd.print("Temp2 00\337C    OK");
  lcd.setCursor(0 + LCD_OFFSET, 2);
  lcd.print("Temp3 00\337C    OK");
  lcd.setCursor(0 + LCD_OFFSET, 3);
  lcd.print("ABGAS 00%     OK");
  return;
}

int pushS1(int heating) {
  if (!digitalRead(SWITCH_1)) {
    while (!digitalRead(SWITCH_1)) delay(100);
    switch (heating) {
      case 0:
        digitalWrite(H_TEMP2, LOW);
        digitalWrite(H_TEMP3, LOW);
        digitalWrite(H_TEMP1, HIGH);
        break;
      case 1:
        digitalWrite(H_TEMP1, LOW);
        digitalWrite(H_TEMP2, HIGH);
        break;
      case 2:
        digitalWrite(H_TEMP2, LOW);
        digitalWrite(H_TEMP3, HIGH);
        break;
      case 3:
        digitalWrite(H_TEMP1, HIGH);
        digitalWrite(H_TEMP2, HIGH);
        digitalWrite(H_TEMP3, HIGH);
        break;
    }
    heating++;
    return (heating > 3 ? 0 : heating);
  }
  return heating;
}

int pushS2(int heating) {
  int resetCMF = 0;
  unsigned long switch_count = 0;
  if (!digitalRead(SWITCH_2)) {
    while (!digitalRead(SWITCH_2)) delay(100);
    switch_count = millis();
    digitalWrite(H_TEMP1, LOW);
    digitalWrite(H_TEMP2, LOW);
    digitalWrite(H_TEMP3, LOW);
    while (!digitalRead(SWITCH_1)) {
      delay(100);
      if (switch_count < millis() - 3000) {
        memStatus = !memStatus;
        switch_count = millis();
        lcd.setCursor(10, 3);
        lcd.print(memStatus ? '_' : ' ');
        resetCMF = 0xFF;
      }
    }
    if (resetCMF) {
      analogWrite(L_TEMP, 0);
      analogWrite(L_GAS, 0);
      status = 0;
    }
    return 0;
  }
  return heating;
}

int receive_sensor(int opt) {
  int value = 0x00;
  value = (opt < 3 ? (int)get_LM75_temperature(opt, 0) : (int)map(analogRead(A0), 35, 870, 0, 100));
  delay(100);
  lcd.setCursor(6 + (opt > 1 ? LCD_OFFSET : 0), opt);
  if (value < 10) lcd.print('0');
  lcd.print(value);
  lcd.setCursor(11 + (opt > 1 ? LCD_OFFSET : 0), opt);
  if (opt < 3 && (value >= oTEMP || value <= uTEMP)) {
    if (value >= oTEMP && !bitRead(status, opt)) {
      lcd.print("ALARM");
      analogWrite(L_TEMP, (memStatus ? 255 : 0));
      bitWrite(status, opt, 1);
      if (!bitRead(status, 5)) {
        toChange = true;
        bitSet(status, 5);
      }
    } else if (value <= uTEMP && bitRead(status, opt)) {
      lcd.print("   OK");
      bitWrite(status, opt, 0);
      if (!(status & 0b00001111)) {
        toChange = true;
        if (!bitRead(status, 4)) analogWrite(L_TEMP, 0);
        bitClear(status, 5);
      }
    }
  } else if (opt == 3) {
    if (value >= oGAS && !bitRead(status, 4)) {
      lcd.print("ALARM");
      bitSet(status, 4);
      bitClear(status, 3);
      analogWrite(L_GAS, (memStatus ? 255  : 0));  // PWM Ansteuerung
      //digitalWrite(L_GAS, (memStatus ? HIGH : LOW));
      analogWrite(L_TEMP, (memStatus ? 255 : 0));
      toChange = true;
    } else if (value >= mGAS && value < oGAS && !bitRead(status, 3)) {
      analogWrite(L_GAS, (memStatus ? 255 : 0));  // PWM Ansteuerung
      delay(1000);
      analogWrite(L_GAS, (memStatus ? 190 : 0));  // PWM Ansteuerung
      //digitalWrite(L_GAS, (memStatus ? HIGH : LOW));
      if (bitRead(status, 4)) {
        timestamp = (!timestamp ? millis() : timestamp);
        if (timestamp < millis() - 5000) {
          if (!(status & 0b00000111))
            analogWrite(L_TEMP, 0);
          timestamp = 0;
          bitClear(status, 4);
          toChange = true;
        } else
          return;
      }
      lcd.print("     ");
      bitSet(status, 3);
      if (!bitRead(status, 5)) {
        bitSet(status, 5);
        toChange = true;
      }
    } else if (value <= uGAS && (bitRead(status, 3) || bitClear(status, 4))) {
      lcd.print("   OK");
      if (bitRead(status, 4)) {
        bitClear(status, 4);
        bitSet(status, 3);
      }
      timestampgas = (!timestampgas ? millis() : timestampgas);
      if (timestampgas < millis() - 5000) {
        bitClear(status, 3);
        analogWrite(L_GAS, 0);  // PWM Ansteuerung
        //digitalWrite(L_GAS, LOW);
        timestampgas = 0;
      }
      if (!(status & 0b00000111)) {
        bitClear(status, 5);
        analogWrite(L_TEMP, 0);
        toChange = true;
      }
    }
  }
  return;
}

void sendDigitalValue(byte ADDR, byte value) {
  Wire.beginTransmission(ADDR);
  Wire.write(value);
  Wire.endTransmission();
}

void signalGoGreen(int opt) {
  int signals[5][3] = { { 2, 3, 0 }, { 5, 6, 0 }, { 11, 12, 8 }, { 8, 9, 0 }, { 14, 15, 5 } };
  for (int i = 0; i < 3; i++, delay(500)) {
    traffic_signal(signals[opt][i]);
  }
  return;
}

void signalGoRed(int opt) {
  int signals[5][3] = { { 0, 1, 2 }, { 0, 4, 5 }, { 8, 10, 11 }, { 0, 7, 8 }, { 5, 13, 14 } };
  for (int i = 0; i < 3; i++, delay(500)) {
    traffic_signal(signals[opt][i]);
  }
  return;
}

void traffic_signal(int opt) {
  int lights = 0xFF;
  counter_switch = pushS1(pushS2(counter_switch));
  switch (opt) {
    case 0: lights = 0b11011011; break;  // green leds
    case 1: lights = 0b11101101; break;  // yellew leds
    case 2: lights = 0b11110110; break;  // red leds
    case 3: lights = 0b11100100; break;  // red/yellow leds

    case 4: lights = 0b11011101; break;  // west green / east yellow
    case 5: lights = 0b11011110; break;  // west green / east red
    case 6: lights = 0b11011100; break;  // west green / east red/yellow

    case 7: lights = 0b11101011; break;  // west yellow / east green
    case 8: lights = 0b11110011; break;  // west red / east green
    case 9: lights = 0b11100011; break;  // west red/yellow / east green

    case 10: lights = 0b11110101; break;  // west red / east yellow
    case 11: lights = 0b11110110; break;  // west red / east red
    case 12: lights = 0b11110100; break;  // west red / east red/yellow

    case 13: lights = 0b11101110; break;  // west yellow / east red
    case 14: lights = 0b11110110; break;  // west red / east red
    case 15: lights = 0b11100110; break;  // west red/yellow / east red
    default: lights = 0xFF;
  }
  sendDigitalValue(TRAFFICLIGHTS, lights);
  return;
}

// LM75 read temperature. Device = 0-7, regx = TEMP, OBEN, UNTEN (Select register)
double get_LM75_temperature(int device, int regx) {
  int8_t msb;
  int8_t lsb;
  int8_t msb1;
  Wire.beginTransmission(TEMPADDRESS + device);
  Wire.write(regx);
  Wire.endTransmission();
  Wire.beginTransmission(TEMPADDRESS + device);
  Wire.requestFrom(TEMPADDRESS + device, 2);
  if (Wire.available()) {
    msb1 = Wire.read();
    msb = msb1 << 1;  // Remove sign bit, remaining value is now twice as large
    lsb = Wire.read();
  }
  // The highest bit indicates whether 0.5 degrees are added
  lsb = (lsb & 0x80) >> 7;
  Wire.endTransmission();
  if (msb1 < 0x80) {               // Is the value positive?
    return double(msb + lsb) / 2;  // positive value
  } else {
    return double(msb + lsb) / 2 - 128;  // negative value
  }
}

// LM75 set configuration
void set_LM75_config(int device, byte value) {
  Wire.beginTransmission(TEMPADDRESS + device);
  Wire.write(1);  // Select register config
  Wire.write(value);
  Wire.endTransmission();
}

// LM75 read configuration, device = 0-7
byte get_LM75_config(int device) {
  byte reg;
  Wire.beginTransmission(TEMPADDRESS + device);
  Wire.write(1);  // Select register config
  Wire.endTransmission();
  Wire.requestFrom(TEMPADDRESS + device, 1);
  if (Wire.available()) {
    reg = Wire.read();
  }
  Wire.endTransmission();
  return reg;
}

// LM75 set threshold, device = 0-7, regx = Value, Degree as double
void set_LM75_threshold(int device, byte regx, double grad) {
  int answer = 0;
  int8_t msb;
  int8_t lsb = 0;
  uint8_t y = 0;
  boolean neg = false;
  if (grad < 0) {
    msb = abs(int(grad)) + 128;
  } else {
    msb = abs(int(grad));
  }
  if (grad - abs(int(grad)) > 0) {
    lsb = 0x80;
  }
  Wire.beginTransmission(TEMPADDRESS + device);
  Wire.write(regx);  // select upper or lower register
  Wire.write(msb);
  Wire.write(lsb);
  answer = Wire.endTransmission();
}
